/* GNU Mailutils -- a suite of utilities for electronic mail
   Copyright (C) 1999, 2000, 2001, 2002, 2005, 2006 Free Software Foundation, Inc.

   GNU Mailutils is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   GNU Mailutils is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with GNU Mailutils; if not, write to the Free Software
   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
   MA 02110-1301 USA */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <string.h>
#ifdef HAVE_STRINGS_H
# include <strings.h>
#endif
#include <errno.h>
#include <mailutils/error.h>
#include <mailutils/errno.h>
#include <mu_dbm.h>

int
mu_fcheck_perm (int fd, int mode)
{
  struct stat st;

  if (fstat (fd, &st) == -1)
    {
      if (errno == ENOENT)
	return 0;
      else
	return 1;
    }
  if ((st.st_mode & 0777) != mode)
    {
      errno = MU_ERR_UNSAFE_PERMS;
      return 1;
    }
  return 0;
}

int
mu_check_perm (const char *name, int mode)
{
  struct stat st;

  if (mode == 0)
    return 0;
  if (stat (name, &st) == -1)
    {
      if (errno == ENOENT)
	return 0;
      else
	return 1;
    }
  if ((st.st_mode & 0777) != mode)
    {
      errno = MU_ERR_UNSAFE_PERMS;
      return 1;
    }
  return 0;
}

#if defined(WITH_GDBM)

int
mu_dbm_open (char *name, DBM_FILE *db, int flags, int mode)
{
  int rc;
  int f;
	  
  if (mu_check_perm (name, mode))
    return -1;

  switch (flags)
    {
    case MU_STREAM_CREAT:
      f = GDBM_NEWDB;
      break;
    case MU_STREAM_READ:
      f = GDBM_READER;
      break;
    case MU_STREAM_RDWR:
      f = GDBM_WRCREAT;
      break;
    default:
      errno = EINVAL;
      return 1;
    }
  db->db = gdbm_open(name, 512, f|GDBM_NOLOCK, mode, NULL);
  rc = db->db == NULL;
  if (rc == 0)
    db->name = strdup (name);
  db->locker = NULL;
  return rc;
}

int
mu_dbm_close (DBM_FILE *db)
{
  gdbm_close(db->db);
  free (db->name);
  return 0;
}

int
mu_dbm_fetch (DBM_FILE *db, DBM_DATUM key, DBM_DATUM *ret)
{
  *ret = gdbm_fetch(db->db, key);
  return ret->dptr == NULL;
}

int
mu_dbm_delete (DBM_FILE *db, DBM_DATUM key)
{
  return gdbm_delete (db->db, key);
}

int
mu_dbm_insert (DBM_FILE *db, DBM_DATUM key, DBM_DATUM contents, int replace)
{
  return gdbm_store(db->db, key, contents,
		    replace ? GDBM_REPLACE : GDBM_INSERT);
}

DBM_DATUM
mu_dbm_firstkey (DBM_FILE *db)
{
  return gdbm_firstkey (db->db);
}

DBM_DATUM
mu_dbm_nextkey (DBM_FILE *db, DBM_DATUM key)
{
  return gdbm_nextkey (db->db, key);
}

void
mu_dbm_datum_free(DBM_DATUM *datum)
{
	void *ptr = MU_DATUM_PTR(*datum);
	if (ptr)
		free(ptr);
	MU_DATUM_PTR(*datum) = 0;
}

#elif defined(WITH_BDB)

int
mu_dbm_open (char *name, DBM_FILE *dbm, int flags, int mode)
{
  int f, rc;
  DB *db;

  if (mu_check_perm (name, mode))
    return -1;

  switch (flags)
    {
    case MU_STREAM_CREAT:
      f = DB_CREATE|DB_TRUNCATE;
      break;
    case MU_STREAM_READ:
      f = DB_RDONLY;
      break;
    case MU_STREAM_RDWR:
      f = DB_CREATE;
      break;
    default:
      errno = EINVAL;
      return -1;
    }

#if DB_VERSION_MAJOR == 2
  rc = db_open (name, DB_HASH, f, mode, NULL, NULL, &db);
#else
  rc = db_create (&db, NULL, 0);
  if (rc != 0 || db == NULL)
    return rc;
# if DB_VERSION_MAJOR == 3
  rc = db->open (db, name, NULL, DB_HASH, f, mode);
# else
  rc = db->open (db, NULL, name, NULL, DB_HASH, f, mode);
# endif
#endif
  if (rc)
    return -1;
  
  dbm->name = strdup (name);
  dbm->locker = NULL;
  dbm->db = db;
  dbm->dbc = NULL;
  return 0;
}

int
mu_dbm_close (DBM_FILE *db)
{
  db->db->close (db->db, 0);
  free (db->name);
  return 0;
}

int
mu_dbm_fetch (DBM_FILE *db, DBM_DATUM key, DBM_DATUM *ret)
{
  return db->db->get (db->db, NULL, &key, ret, 0);
}

int
mu_dbm_delete (DBM_FILE *db, DBM_DATUM key)
{
  return db->db->del (db->db, NULL, &key, 0);
}

int
mu_dbm_insert (DBM_FILE *db, DBM_DATUM key, DBM_DATUM contents, int replace)
{
  /*FIXME: replace unused*/
  return db->db->put (db->db, NULL, &key, &contents, 0);
}

DBM_DATUM
mu_dbm_firstkey (DBM_FILE *db)
{
  DBT key, data;
  int ret;

  memset(&key, 0, sizeof key);
  memset(&data, 0, sizeof data);

  if (!db->dbc)
    {
      if (db->db->cursor(db->db, NULL, &db->dbc BDB2_CURSOR_LASTARG) != 0)
	return key;
    }

  if ((ret = db->dbc->c_get(db->dbc, &key, &data, DB_FIRST)) != 0)
    {
      key.data = NULL;
      key.size = 0;
      if (ret == DB_NOTFOUND)
	errno = MU_ERR_NOENT;
      else
	errno = ret;
    }
  return key;
}

DBM_DATUM
mu_dbm_nextkey (DBM_FILE *db, DBM_DATUM pkey /*unused*/)
{
  DBT key, data;
  int ret;

  memset(&key, 0, sizeof key);
  memset(&data, 0, sizeof data);

  if (!db->dbc)
    return key;

  if ((ret = db->dbc->c_get(db->dbc, &key, &data, DB_NEXT)) != 0)
    {
      key.data = NULL;
      key.size = 0;
      if (ret == DB_NOTFOUND)
	errno = MU_ERR_NOENT;
      else
	errno = ret;
    }
  return key;
}

void
mu_dbm_datum_free(DBM_DATUM *datum)
{
	/* empty */
}

#endif

#ifdef USE_DBM

int
mu_dbm_lock (DBM_FILE *db)
{
  int rc;

  if (db->locker)
    return 0;
  
  if ((rc = mu_locker_create (&db->locker, db->name, 0)))
    {
      mu_error ("cannot create locker for %s: %s", db->name, mu_strerror (rc));
      return 1;
    }

  if ((rc = mu_locker_lock (db->locker)))
    {
      mu_locker_destroy (&db->locker);
      return rc;
    }
  return 0;
}
  
void
mu_dbm_unlock (DBM_FILE *db)
{
  mu_locker_unlock (db->locker);
  mu_locker_destroy (&db->locker);
}
  

#endif
